from WidgetFinder import NodeWidgetFinder
from EventFinder import EventFinder

class EventManager:
    """Mixin for ZWidgetManager dealing with events.
    """
    
    def _key(self, document, klass):
        """Create a session key for a document and class.
        FIXME: this would cause a bug if two classes have the same name.
        """
        return klass.__name__ + document._p_oid
        
    def _get(self, document, klass):
        """Get an object of a certain class for a certain document.
        """
        if not hasattr(self, 'FSSession'):
            raise Error, "Event system for XMLWidgets requires FSSession."
    
        session = self.FSSession
        
        key = self._key(document, klass)

        if session.has_key(key):
            return session[key]
        else:
            obj = klass()
            session.set(key, obj)
            return obj

    def _set(self, document, obj):
        if not hasattr(self, 'FSSession'):
            raise Error, "Event system for XMLWidgets requires FSSession."

        session = self.FSSession

        key = self._key(document, obj.__class__)
    
        session.set(key, obj)

    def clearSession(self):
        """Clear everything in session.

        NOTE: Session shouldn't be shared by sessionID with other data or
        that will be removed as well!
        """
        if not hasattr(self, 'FSSession'):
            raise Error, "Event system for XMLWidgets requires FSSession."

        session = self.FSSession
        # remove everything in session
        for key in session.keys():
            del session[key]
        
    def addNodeWidget(self, widget_id, node):
        """Add a node widget to the nodewidgetfinder.

        widget_id -- the id of the widget that will handle node
        node -- the node to use the special widget for
        """
        # get owner document of node
        document = node.getOwnerDocument()
        # get nodewidgetfinder
        nodewidgetfinder = self._get(document, NodeWidgetFinder)

        # add node widget
        widget = getattr(self, widget_id)
        nodewidgetfinder.addWidget(widget, node)

        # store session data again
        self._set(document, nodewidgetfinder)

    def clearNodeWidgets(self, document):
        """Clear all node-specific widgets for document from session.
        """
        # get information from session
        nodewidgetfinder = self._get(document, NodeWidgetFinder)

        # add node widget
        nodewidgetfinder.clear()

        # store session data again
        self._set(document, nodewidgetfinder)
        
    def _findNodeWidgetId(self, node):
        """Look up if node occurs in NodeWidgetFinder for this document.

        Result: widget_id if widget found, or None if none found
        """
        nodewidgetfinder = self._get(node.getOwnerDocument(),
                                     NodeWidgetFinder)
        return nodewidgetfinder.find(node)

    def _get_ids(self, node, widget, methodname):
        """Get display_id and event_id for widget's method on node.

        node -- the node to add the event for
        widget -- the widget that will handle the event
        methodname -- the name of the event (on the widget)
        """
        # get owner document of node
        document = node.getOwnerDocument()
        # get eventfinder from session
        eventfinder = self._get(document, EventFinder)

        # add event to eventfinder
        event_id = eventfinder.addEvent(node, widget, methodname)
        
        # store eventfinder in session again
        self._set(document, eventfinder)

        # return display_id and event_id
        return eventfinder.display_id, event_id

    def _base_url(self):
        """Get base of event url.
        """
        if self.viewMethod:
            return self.viewmethod
        else:
            return "."
        
    def event_hyperlink_url(self, node, widget, methodname):
        """Create URL for hyperlink event handler.

        node -- the node to add the event for
        widget -- the widget that will handle the event
        methodname -- the name of the event (on the widget)
        """
        display_id, event_id = self._get_ids(node, widget, methodname)

        # construct URL to return
        
        # create URL arguments
        arguments = "?display_id=%s&event_id=%s" % (display_id, event_id)

        # construct URL
        return self._base_url() + arguments

    def event_form_url(self):
        """Create URL for form event handler.

        NOTE: this url is used by form action events. Uses a special
        redirect page if available (necessary for #focus operations).
        """
        return self._base_url()
  
    def event_form_hidden(self, node, widget, methodname):
        """Create hidden field for form with event_id.
        """
        display_id, event_id = self._get_ids(node, widget, methodname)

        # construct hidden fields we always need
        template = '<input type="hidden" name="%s" value="%s">\n'
        display_field = template % ('display_id', display_id)
        event_field = template % ('event_id', event_id)
        hidden_fields = display_field + event_field

        # return hidden fields
        return hidden_fields

    def clearEvents(self, document):
        """Clear events for document from eventfinder.
        """    
        eventfinder = self._get(document, EventFinder)
        
        eventfinder.clear() # also updates display_id
    
        self._set(document, eventfinder)

    def handleEvent(self, document, display_id, event_id, REQUEST):
        """Handle incoming event.

        document -- the document for which the event is arriving
        display_id -- the display_id associated with the event_id
                      (has to be current, otherwise we'll get an error)
        event_id -- the id of the event
        REQUEST -- web request object
        """
        eventfinder = self._get(document, EventFinder)

        display_id = int(display_id)
        if display_id != eventfinder.display_id:
            raise BackError, "display_id %s is not current. Expected: %s" %\
                  (display_id, eventfinder.display_id)
        
        # look up event
        node_url, widget_id, methodname = eventfinder.find(int(event_id))

        # find node associated with url
        node = REQUEST.resolve_url(node_url)

        # execute methodname on widget in context
        return self.callMethodInContext(node, getattr(self, widget_id),
                                        methodname, REQUEST)
        
class Error(Exception):
    pass

class BackError(Exception):
    pass
