from Globals import HTMLFile
from Globals import MessageDialog
from Globals import Persistent

import OFS.SimpleItem            
import Acquisition               
import AccessControl.Role      

import string
from cStringIO import StringIO

from OFS import ZDOM

from WidgetFinder import NodeNameWidgetFinder
from EventManager import EventManager

class Error(Exception):
    pass

class ZWidgetManager(
    OFS.SimpleItem.Item,
    Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    EventManager
    ):

    meta_type = "XMLWidgetManager"

    # the various tabs
    manage_options = (
        {'label' : 'Edit',     'action' : 'manage_main'},
        {'label' : 'Security', 'action' : 'manage_access'},
        )

    # display edit screen as main management screen
    manage_main = HTMLFile('widgetmanagerEdit', globals())

    # permissions
    __ac_permissions__ = (
        ('View management screens',
         ('manage_tabs',
          'manage_main',)),

        ('Change permissions',
         ('manage_access',)),

        ('Change XMLWidgetManagers',
         ('manage_edit',)),
          
        ('View',        
         ('find',
          'processAll',)),
        )

    def __init__(self, id, title):
        """Create a new widget manager.
        """
        self.id = id
        self.title = title
        self.widgetnames = []
        self.widgetfinder = NodeNameWidgetFinder()
        self.defaultWidget = None
        self.viewMethod = None
        self.redirectMethod = None
        
    def manage_edit(self, title, REQUEST=None):
        """Handle result from edit action.
        title - the new title of the manager
        REQUEST - contains other data
        """
        # set title
        self.title = title
                
        if REQUEST is not None:
            # extract list of widget names
            widgetnames = map(string.strip,
                              string.split(REQUEST['widgets'], "\n"))
            # filter out all empty names
            widgetnames = filter(lambda name: name != "", widgetnames)

            # set widgets in widgetfinder 
            self._setWidgets(widgetnames)

            # set default widget
            defaultWidgetName = string.strip(REQUEST.defaultWidget)
            if defaultWidgetName != "":
                # set default widget to what we find
                self.defaultWidget = self._getWidget(defaultWidgetName)
            else:
                self.defaultWidget = None
                
            # set view method
            viewMethod = string.strip(REQUEST.viewMethod)
            if viewMethod != "":
                self.viewMethod = viewMethod
            else:
                self.viewMethod = None

                
        # done with editing
        if REQUEST is not None:
            message = "Properties changed"
            return self.manage_main(self, REQUEST,
                                    manage_tabs_message=message)
        
    def _setWidgets(self, widgetnames):
        """Set widgets in the widgetfinder.
        """
        # clear out previous widgets
        self._clearWidgets()
    
        # store the names in this object
        self.widgetnames = widgetnames

        widgetfinder = self.widgetfinder
        
        for widgetname in widgetnames:
            # get widget with this name
            widget = self._getWidget(widgetname)
            # add it to the widgetfinder
            widgetfinder.addWidget(widget)

        self.widgetfinder = widgetfinder
        
    def _getWidget(self, widget_id):
        """Get widget with widget_id.
        FIXME: this is a hack to make it work with ZClasses
        """
        # get meta_type to determine if we're in a ZClass or not
        try:
            meta_type = self.aq_parent.meta_type()
        except TypeError:
            meta_type = self.aq_parent.meta_type
            
        if meta_type != 'Z Class':
            # not in a Z Class, in a normal folder
            return getattr(self, widget_id)
        else:
            # we are in a Z Class
            return self.aq_parent._getOb(widget_id)
            
    def _clearWidgets(self):
        """Clear all widgets.
        """
        # remove widgetnames from this object
        self.widgetnames = []
        # clear all widgets from widgetfinder
        self.widgetfinder.clear()
        self.widgetfinder = self.widgetfinder

    # FIXME: is this still necessary?
    index_html = None
    
    def find(self, node, REQUEST):
        """Find the widget to draw with.
        """
        # no widget found yet
        widget_id = None
        
        # nodewidgetfinder is not supported if FSSession isn't available
        if hasattr(self, 'FSSession'):
            # try to look up widget_id in NodeWidgetFinder (in session)
            widget_id = self._findNodeWidgetId(node)

        # if we haven't found a widget yet,
        # look up the widget for this nodename
        if not widget_id:
            widget_id = self.widgetfinder.find(node, REQUEST)
            
        # if we found the widget id
        if widget_id:
            # get and return widget
            return getattr(self, widget_id)
        else:
            # use default widget if available and if we're
            # looking up an element node
	    if self.defaultWidget and node.getNodeType() == ZDOM.ELEMENT_NODE:
                return self.defaultWidget
	    # otherwise, prepare hopefully informative error
            errtable = {
                'nodename': node.getNodeName(),
                'parentname': node.getParentNode().getNodeName()
                }
            
            errstr = "Could not find XMLWidget for '%(nodename)s' node in "\
                     "context of '%(parentname)s' node." % errtable

            # raise error
            raise Error, errstr
    
    def callMethodInContext(self, node, widget, methodname, REQUEST):
        """Magic acquisition voodoo.

        Call a method on widget in the context of:

          * the folder/ZClass everything resides in
          
          * any parent widget of the widget (inheritance)

          * the widget itself

          * the node
        """
        # get widget in inheritance/acquisition context
        widget = widget._inheritContext()
        # set widget in REQUEST to we can refer to it in widget methods
        # FIXME: this shouldn't be necessary anymore
        REQUEST.set("xml_widget", widget)

        # get node in acquisition context of widget
        parent_nodes = []
        parent = node.getParentNode()
        while parent:
            parent_nodes.append(parent)
            parent = parent.getParentNode()
        parent_nodes.reverse()

        node = node.__of__(widget)
        for parent_node in parent_nodes:
            node = node.__of__(parent_node)
                
        # now get method in widget, call it in node context, and return result
        return getattr(widget, methodname)(node, REQUEST)

    def process(self, node, methodname, REQUEST):
        """Call methodname on node for associated widget and return result.
        """
        return self.callMethodInContext(node, self.find(node, REQUEST),
                                         methodname, REQUEST)
    
    def processAll(self, parentnode, methodname, REQUEST):
        """Call methodname on all subnodes and return result.
        """
        result = StringIO()
        for node in parentnode.getChildNodes():
            result.write(self.process(node, methodname, REQUEST)) 
        return result.getvalue()

    def qd_process(self, node, methodname, REQUEST):
        """Quick & dirty variety of process().

        - Does not support widget inheritance.
        - Does not support calling other methods on the widget
          directly from a widget method (use xml_widget instead).
        - should run more quickly.
        """
        return getattr(self.find(node, REQUEST), methodname)(node, REQUEST)

    def qd_processAll(self, parentnode, methodname, REQUEST):
        """Quick & dirty variety of processAll().
        """
        result = StringIO()
        for node in parentnode.getChildNodes():
            result.write(self.process(node, methodname, REQUEST))
        return result.getvalue()
        
# form to add the WidgetManager to a folder
manage_addWidgetManagerForm = HTMLFile('widgetmanagerAdd', globals())

# adds WidgetManager to folder
def manage_addWidgetManager(self, id, title='', REQUEST=None):
    """Add manager to folder.
    """
    wm = ZWidgetManager(id, title)
    
    self._setObject(id, wm)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

